/*
 * ChickenMobile.c
 *
 * Created: 09.10.2016 17:42:05
 * Author : SLW
 */ 

#ifndef cmh_f_cpu_
   #define cmh_f_cpu_
   #define F_CPU 8000000L
#endif

#include <inttypes.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "chicken_mobile.h"
#include "DS18B20_routines.h"
#include "glcd_routines.h"
#include "i2cmaster.h"
#include "keyboard_routines.h"
#include "display_routines.h"
#include "cmh_util.h"
#include "cmh_menus.h"
#include "ee_routines.h"

// date and time of the compiler run
const char timeP[] PROGMEM = __TIME__;
const char dateP[] PROGMEM = __DATE__;

// global system parameter
display_status_t display_status;
datetime_t date_time;						// current date and time, filled from RTC 
temp_status_t temp_status;
amb_light_status_t amb_light_status;	
batt_status_t batt_status;
LED_status_t LED_status;
mp3_status_t mp3_status;
hatch_control_t hatch_control;

// global program and timing parameters
volatile uint8_t sec_counter;				// counts seconds (0 ... 59)
volatile uint8_t job_flags;					// interrupt driven job flags
volatile uint8_t scheduler_flags;			// interrupt driven scheduled flags
volatile uint8_t mp3_trigger;				// manages 300msec impulse for mp3 player

// log file buffer
int16_t temp_buffer_ext[TEMP_BUFFER_SIZE];
int16_t temp_buffer_int[TEMP_BUFFER_SIZE];
uint16_t battery_buffer[TEMP_BUFFER_SIZE];
uint8_t temp_buffer_pnt;
light_start_time_t light_start_time[LIGHT_START_BUFFER_SIZE];
uint8_t light_start_pnt;

/* run_scheduler -------------------------------------------------------------------------
* Checks scheduler flags for pending actions and executes those
* To be used while the system is waiting for key entries
*/
void run_scheduler(void) {
	
	// run once-per-second jobs
	if (job_flags & (1 << JF_ONCE_PER_SECOND)) {
		job_flags &= ~(1 << JF_ONCE_PER_SECOND);
		mp3_sequencer();
		hatch_sequencer();
	};
	
	// anything else to do?
	if (scheduler_flags == 0) return;					// nothing to do
	
	if (scheduler_flags & (1 << SF_SYNC_TIME)) {
		scheduler_flags &= ~(1 << SF_SYNC_TIME);
		get_date_time();
		show_time(&date_time, DISPLAY_POS_TIME_X, 0);
	};
	if (scheduler_flags & (1 << SF_READ_EXT_TEMP)) {
		scheduler_flags &= ~(1 << SF_READ_EXT_TEMP);
		temp_status.err_flag_ext = ds18b20_read_temperatur(DS18B20_PIN_EXT, &temp_status.temp_ext);
	};
	if (scheduler_flags & (1 << SF_READ_INT_TEMP)) {
		scheduler_flags &= ~(1 << SF_READ_INT_TEMP);
		temp_status.err_flag_int = ds18b20_read_temperatur(DS18B20_PIN_INT, &temp_status.temp_int);
		log_temperature();
	};
	if (scheduler_flags & (1 << SF_CHECK_BATT_STATUS)) {
		scheduler_flags &= ~(1 << SF_CHECK_BATT_STATUS);
		check_batt_status();
	};
	if (scheduler_flags & (1 << SF_CHECK_LIGHT_STATUS)) {
		scheduler_flags &= ~(1 << SF_CHECK_LIGHT_STATUS);
		check_amb_light_status();
	};
	if (scheduler_flags & (1 << SF_UPDATE_LED_STATUS)) {
		scheduler_flags &= ~(1 << SF_UPDATE_LED_STATUS);
		update_LED_status();
	};
	if (scheduler_flags & (1 << SF_UPDATE_MP3_STATUS)) {
		scheduler_flags &= ~(1 << SF_UPDATE_MP3_STATUS);
		update_mp3_status();
	};
	if (scheduler_flags & (1 << SF_CHECK_HATCH)) {
		scheduler_flags &= ~(1 << SF_CHECK_HATCH);
		update_hatch_status();
	};
}

/* ---------------------------------------------------------------------------------------*/
int main(void)
{
	extern const char str_init_master[];
	extern const char str_init_version[];
	extern const char str_init_keyboard[];
	extern const char str_init_bat_status[];
	extern const char str_init_amb_light_status[];
	extern const char str_init_LED_status[];
	extern const char str_init_temp_sensors[];
	extern const char str_init_RTC[];
	extern const char err_ext_temp_sensor[];
	extern const char err_int_temp_sensor[];
	extern const char str_init_mp3_status[];
	extern const char str_init_hatch_control[];

	uint8_t keyboard_value, display_level = 0;
	const uint8_t y_step = 10;
	uint8_t x_pos = 0, y_pos = 0;

	// Set port direction
	DDR_MP3 |= (PORTBIT_MP3_ON | PORTBIT_MP3_SONG);
	DDR_POWER |= (PORTBIT_POWER);
	DDRC = 0b11110000;						// ADC: set ports to input
	DDRD |= (1 << DDD6);					// LED PWM: set PD6 to output
	DDR_HATCH |= (1 << PORTBIT_HATCH);		// hatch control is output
	hatch_switch_off();						// initial value is off = high

	// set initial output values	
	PORT_POWER |= (1 << PORTBIT_POWER);
	switch_mp3_off();
	PORT_MP3 &= ~(1 << PORTBIT_MP3_SONG);

	// initialize ADC 
	ADMUX  = (1<<REFS0) | (1<<REFS1) | (0<<ADLAR);					// 10 bit, right adjusted, internal reference
	ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (0<<ADPS0);		// scaling factor: 64 -> 12500 kHz
	// start first conversion
	ADMUX  = (1<<REFS1) | (0<<ADLAR) | 1;							// set channel
	ADCSRA |= (1<<ADSC);											// start first ADC conversion

	// initiate TWM
	_delay_ms(300);
	i2c_init();														// initiate TWR interface with 200 kHz
	_delay_ms(500);
	
	// initiate display driver
	ee_read_display_status();										// display parameters
	glcd_set_verbose(255);
	glcd_set_font(3);
	glcd_clear_screen();
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_master);
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_version);
	glcd_draw_pgm_string(dateP);
	_delay_ms(500);

	// initiate keyboard
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_keyboard);
	init_keyboard();
	_delay_ms(500);
	
	// initiate status values and data from ee_prom
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_bat_status);
	ee_read_batt_status();
	batt_status.vcc_status = BATT_LEVEL_HIGH;
	_delay_ms(500);
		
	// initiate amb-light status
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_amb_light_status);
	ee_read_amb_light_status();										// ambient light parameters
	amb_light_status.light_mov_avg = 2000;							// initially invalid
	amb_light_status.sens_status = LIGHT_LEVEL_HIGH;				// initial status
	_delay_ms(500);
	
	// initiate LED status
	x_pos = 110;
	y_pos = 0;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_LED_status);
	ee_read_LED_status();											// LED parameters
	if (time_after(date_time.hour, date_time.min, LED_status.led_tw_off_hour, LED_status.led_tw_off_min) > 0)
		LED_status.led_status = LED_FIN;							// no LED for the rest of the day
	else
		LED_status.led_status = LED_OFF;							// LED off for now, but maybe switched on later
	set_LED_level(0);
	for (light_start_pnt = 0; light_start_pnt < LIGHT_START_BUFFER_SIZE; ++light_start_pnt) {
		light_start_time[light_start_pnt].hour = 0xff;
		light_start_time[light_start_pnt].min = 0xff;
	};
	light_start_pnt = 0;
	_delay_ms(500);
		
	// initiate DS18B20 temperature sensors
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_temp_sensors);
	if ((temp_status.err_flag_ext = ds18b20_init(DS18B20_PIN_EXT)) == 0) {
		temp_status.err_flag_ext = ds18b20_read_temperatur(DS18B20_PIN_EXT, &temp_status.temp_ext);
	} else {
		glcd_draw_pgm_string(err_ext_temp_sensor);
		_delay_ms(1000);
	};
	if ((temp_status.err_flag_int = ds18b20_init(DS18B20_PIN_INT)) == 0) {
		temp_status.err_flag_int = ds18b20_read_temperatur(DS18B20_PIN_INT, &temp_status.temp_int);
	} else {
		glcd_draw_pgm_string(err_int_temp_sensor);
		_delay_ms(1000);
	};
	// clear temperature and battery voltage buffer
	for (temp_buffer_pnt = 0; temp_buffer_pnt < TEMP_BUFFER_SIZE; ++temp_buffer_pnt) {
	 	temp_buffer_ext[temp_buffer_pnt] = MAX_TEMP;
		temp_buffer_int[temp_buffer_pnt] = MAX_TEMP;
		battery_buffer[temp_buffer_pnt] = MAX_BATT;
	};
	temp_buffer_pnt = 0;
	
	_delay_ms(500);	
	
	// initiate date and time
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_RTC);
	get_date_time();
	_delay_ms(500);
	
	// initiate MP3 status
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_mp3_status);
	ee_read_mp3_status();	
	mp3_status.mp3_tw1_status = MP3_TW_OFF;
	mp3_status.mp3_tw2_status = MP3_TW_OFF;
	mp3_status.song_nr = 1;
	_delay_ms(500);

	// initiate hatch control
	y_pos += y_step;
	glcd_set_cursor(x_pos, y_pos);
	glcd_draw_pgm_string(str_init_hatch_control);
	ee_read_hatch_control();
	hatch_control.hatch_cnt = 0;
	_delay_ms(500);
	
	// initiating cockpit display
	display_startup();
	_delay_ms(1000);
	glcd_clear_screen();
	_delay_ms(5);
	initiate_cockpit();
	display_status.timeout_cnt = display_status.timeout;

	// set sleep mode
	set_sleep_mode(SLEEP_MODE_IDLE);

	// Configure timer1 for interrupts at 20 Hz (8000000 / 64 / 6250 -> 20)
	OCR1A = 6250-1;
	TCCR1B |= (1<<WGM12)| (0<<CS02) | (1<<CS01) |(1<<CS00);  		// CPU speed / 64, CTC1 = Clear Timer on Compare
	TIMSK1 |= (1 << OCIE1A);
	sec_counter = 0;
	sei();
	
	while (1) {
		
		if ((display_level > 0) && (display_status.timeout_cnt == 0)) keyboard_value = 1;	// auto cancel when display switches off
		else keyboard_value = read_keyboard();						// otherwise check whether key was pressed
		
		switch (keyboard_value) {
			case 1:
				display_level = 0;
				break;
			case 3:			// test output
				run_test_output();
				break;
			case 4:			// setup menu
				run_setup_menu();
				break;
			case 5:
			case 7:
				if (display_level < 4)
					++display_level;
				else
					display_level = 0;
				break;
			case 6:
				if (display_level > 0) 
					--display_level;
				else
					display_level = 4;
				break;
		};


		if (keyboard_value > 0) {
			switch (display_level) {
				case 0:
					initiate_cockpit();
					break;
				case 1:
					show_temp_log_int();
					break;
				case 2:
					show_temp_log_ext();
					break;
				case 3:
					show_batt_log();
					break;
				case 4:
					show_light_start_time();
					break;
			};
		};
		
		if (display_level == 0) {
			if (job_flags & (1 << JF_REFRESH_DISPLAY)) {
				job_flags &= ~(1 << JF_REFRESH_DISPLAY);
				refresh_cockpit();
			};
		};
		
		if (job_flags & (1 << JF_DISPLAY_LIGHT_ON)) {				// switch light on
			job_flags &= ~(1 << JF_DISPLAY_LIGHT_ON);
			glcd_clear_dim();
		};
		if (job_flags & (1 << JF_DISPLAY_LIGHT_OFF)) {				// switch light off
			job_flags &= ~(1 << JF_DISPLAY_LIGHT_OFF);
			glcd_set_dim();
		};
    };
}

/* Interrupt Service Handler ---------------------------------------------------------------------*/
/* Runs 20 times per second 														              */
ISR(TIMER1_COMPA_vect) {
	static uint8_t irq_counter = 0;					// counts the number of ISR calls, 20 for 1 second
	static uint8_t adc_counter = 0;					// rotates across 4 ADC channels
	static uint16_t adc_vcc = 0;
	static uint16_t adc_sens1 = 0, adc_sens2 = 0;

	// manage ADC
	{
		switch (adc_counter) {
			case 0:
			adc_vcc += ADCW;						// read Vcc conversion value;
			ADMUX  = (1<<REFS1) | (0<<ADLAR) | 1;	// set channel to sens1
			ADCSRA |= (1<<ADSC);					// start ADC conversion
			++adc_counter;
			break;
			case 1:
			// read ADCW channel 1;
			adc_sens1 += ADCW;						// read sens1
			ADMUX  = (1<<REFS1) | (0<<ADLAR) | 2;	// set channel to sens2
			ADCSRA |= (1<<ADSC);					// start ADC conversion
			++adc_counter;
			break;
			case 2:
			// read ADCW channel 2;
			adc_sens2 += ADCW;						// read sens2
			ADMUX  = (1<<REFS1) | (0<<ADLAR) | 3;	// set channel - next, not used
			ADCSRA |= (1<<ADSC);					// start ADC conversion
			++adc_counter;
			break;
			case 3:
			// read ADCW channel 3;
			;										// nut used
			ADMUX  = (1<<REFS1) | (0<<ADLAR) | 0;	// set channel to ADC
			ADCSRA |= (1<<ADSC);					// start ADC conversion
			++adc_counter;
			break;
			case 4:
			adc_counter = 0;
			break;
		};
	};

	// manage MP3 trigger	
	if (mp3_trigger > 0) {
		if (--mp3_trigger == 0) PORT_MP3 &= ~(1 << PORTBIT_MP3_SONG);
	};
	
	// check keyboard
	{
		check_keyboard();
	};
	
	// once per second: summarize data and manage job and scheduler flags
	{
		++irq_counter;
		if (irq_counter == 20) {
			irq_counter = 0;
			batt_status.vcc = adc_vcc * 6 / 16;
			adc_vcc = 0;
			amb_light_status.sens1 = adc_sens1 / 4;
			adc_sens1 = 0;
			amb_light_status.sens2 = adc_sens2 / 4;
			adc_sens2 = 0;
			
			if (display_status.timeout_cnt > 0) {
				--display_status.timeout_cnt;
				if (display_status.timeout_cnt == 0) job_flags |= (1 << JF_DISPLAY_LIGHT_OFF);
			};

			job_flags |= (1 << JF_ONCE_PER_SECOND);				// once per second
			job_flags |= (1 << JF_REFRESH_DISPLAY);				// refresh display 
			if (++sec_counter == 60) sec_counter = 0;
			
			// set scheduler flags
			switch (sec_counter) {
				case 0:
					scheduler_flags |= (1 << SF_SYNC_TIME);
					break;
				case 5:
					scheduler_flags |= (1 << SF_UPDATE_MP3_STATUS);
					break;
				case 10:
					scheduler_flags |= (1 << SF_READ_EXT_TEMP);
					break;
				case 15:
					scheduler_flags |= (1 << SF_CHECK_HATCH);
					break;
				case 20:
					scheduler_flags |= (1 << SF_READ_INT_TEMP);
					break;
				case 30:
					scheduler_flags |= (1 << SF_CHECK_BATT_STATUS);
					break;
				case 40:
					scheduler_flags |= (1 << SF_CHECK_LIGHT_STATUS);
					break;
				case 50:
					scheduler_flags |= (1 << SF_UPDATE_LED_STATUS);
					break;
			};
		};
	};
}


